/**
 * PETALS - PETALS Services Platform. Copyright (c) 2009 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
package org.ow2.petals.ant.task;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;

import org.apache.tools.ant.BuildException;
import org.ow2.petals.ant.AbstractConfigureTask;
import org.ow2.petals.ant.util.ZipUtil;
import org.ow2.petals.ant.util.ZipUtil.ZipEntryCallback;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.ow2.petals.jbi.descriptor.original.JBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.original.generated.Component;
import org.ow2.petals.jbi.descriptor.original.generated.Jbi;

/**
 * Ant task : petals-configure-component. This task pre-configure a JBI
 * component (service engine of binding component).
 * 
 * @author rnaudin
 * 
 */
public class ConfigureComponentTask extends AbstractConfigureTask {

    /**
     * SharedLibrary class.
     * 
     * @author rnaudin
     * 
     */
    public static final class SharedLibrary {

        private String file;

        public SharedLibrary() {
            super();
        }

        public SharedLibrary(String file) {
            super();
            this.file = file;
        }

        public String getFile() {
            return file;
        }

        public void setFile(String file) {
            this.file = file;
        }
    }

    /**
     * Inner structure.
     * 
     * @author rnaudin
     * 
     */
    private static final class _SharedLibrary {
        public String identification;

        public String version;
    }

    private final static String JBIDESCRIPTOR_ZIPENTRY_NAME = "META-INF/jbi.xml";

    /**
     * The localization of the component.
     */
    private String file;

    /**
     * The aimed identification of the component.
     */
    private String identification;

    /**
     * The aimed localization of the result component.
     */
    private String outputFile;

    private final List<SharedLibrary> sharedLibraries = new ArrayList<SharedLibrary>();

    /**
     * Location of the file that contains Shared Libraries properties
     */
    private String slProperties;

    /**
     * A Shared Library to include in the component configuration.
     * 
     * @return
     */
    public SharedLibrary createSharedLibrary() {
        SharedLibrary sharedLibrary = new SharedLibrary();
        this.sharedLibraries.add(sharedLibrary);
        return sharedLibrary;
    }

    @Override
    public void execute() throws BuildException {
        validateFileParameter(this.outputFile, "outputFile");

        final List<_SharedLibrary> _sharedLibraries = new ArrayList<_SharedLibrary>();

        try {
            this.loadSharedLibrariesProperties();

            for (SharedLibrary sharedLibrary : this.sharedLibraries) {
                final URL slUrl = validateFileParameter(sharedLibrary.getFile(),
                        "sharedlibrary.file");
                final File slFile = downloadURL(slUrl);
                final ZipFile slZip = new ZipFile(slFile);
                final ZipEntry jbiZipEntry = slZip.getEntry(JBIDESCRIPTOR_ZIPENTRY_NAME);
                final InputStream jbiInputStream = slZip.getInputStream(jbiZipEntry);
                final Jbi slJbi = JBIDescriptorBuilder.buildJavaJBIDescriptor(jbiInputStream);
                jbiInputStream.close();
                _SharedLibrary _sharedLibrary = new _SharedLibrary();
                _sharedLibrary.identification = slJbi.getSharedLibrary().getIdentification()
                        .getName();
                _sharedLibrary.version = slJbi.getSharedLibrary().getVersion();
                _sharedLibraries.add(_sharedLibrary);
            }

            final URL url = validateFileParameter(this.file, "file");
            final File componentFile = downloadURL(url);
            final ZipOutputStream zipOutputFile = new ZipOutputStream(new FileOutputStream(
                    this.outputFile));

            ZipUtil.copyAndUpdateZipFile(new ZipFile(componentFile), zipOutputFile,
                    new ZipEntryCallback() {
                        public InputStream onZipEntry(final ZipEntry zipEntry,
                                final InputStream zipEntryInputStream) throws IOException,
                                JBIDescriptorException {
                            if (JBIDESCRIPTOR_ZIPENTRY_NAME.equals(zipEntry.getName())) {
                                InputStream resultStream = null;
                                try {
                                    Jbi componentJbi = JBIDescriptorBuilder
                                            .buildJavaJBIDescriptor(zipEntryInputStream);
                                    List<Component.SharedLibrary> componentSLs = componentJbi
                                            .getComponent().getSharedLibraryList();
                                    componentSLs.clear();
                                    for (_SharedLibrary _sharedLibrary : _sharedLibraries) {
                                        Component.SharedLibrary sharedLibrary = new Component.SharedLibrary();
                                        sharedLibrary.setContent(_sharedLibrary.identification);
                                        sharedLibrary.setVersion(_sharedLibrary.version);
                                        componentSLs.add(sharedLibrary);
                                    }

                                    if (ConfigureComponentTask.this.identification != null) {
                                        componentJbi.getComponent().getIdentification().setName(
                                                ConfigureComponentTask.this.identification);
                                    }

                                    String jbiString = JBIDescriptorBuilder
                                            .buildXmlStringJBIdescriptor(componentJbi);

                                    resultStream = new ByteArrayInputStream(jbiString.getBytes());
                                } finally {
                                    zipEntryInputStream.close();
                                }
                                return resultStream;
                            } else {
                                return zipEntryInputStream;
                            }
                        }
                    });
            zipOutputFile.flush();
            zipOutputFile.close();
        } catch (IOException e) {
            throw new BuildException(e);
        } catch (JBIDescriptorException e) {
            throw new BuildException(e);
        }

        log("Component configuration set");
    }

    /**
     * File getter.
     * 
     * @return
     */
    public String getFile() {
        return file;
    }

    /**
     * Identification getter.
     * 
     * @return
     */
    public String getIdentification() {
        return identification;
    }

    /**
     * OutputFile getter.
     * 
     * @return
     */
    public String getOutputFile() {
        return outputFile;
    }

    /**
     * SlProperties getter. 
     * @return
     */
    public String getSlProperties() {
        return slProperties;
    }
    
    /**
     * File setter.
     * 
     * @param file
     */
    public void setFile(String file) {
        this.file = file;
    }

    /**
     * Identification setter.
     * 
     * @param identification
     */
    public void setIdentification(String identification) {
        this.identification = identification;
    }

    /**
     * OutputFile setter.
     * 
     * @param outputFile
     */
    public void setOutputFile(String outputFile) {
        this.outputFile = outputFile;
    }

    /**
     * SlProperties setter.
     * @param slProperties
     */
    public void setSlProperties(String slProperties) {
        this.slProperties = slProperties;
    }

    /**
     * Load the Shared Libraries properties.
     * 
     * @throws IOException
     */
    private void loadSharedLibrariesProperties() throws IOException {
        Properties props = new Properties();
        if (this.slProperties != null) {
            props.load(new FileInputStream(this.slProperties));
        }
        for (Object value : props.values()) {
            SharedLibrary sl = new SharedLibrary((String) value);
            this.sharedLibraries.add(sl);
        }
    }

}
